import { F, O, isObject, isString } from '@eslint-react/eff';
import * as ASTUtils from '@typescript-eslint/utils/ast-utils';
import { AST_NODE_TYPES } from '@typescript-eslint/types';
import * as AST3 from '@eslint-react/ast';
import { DefinitionType, ScopeType } from '@typescript-eslint/scope-manager';
import { isMatching } from 'ts-pattern';

// src/find-variable.ts
var findVariable2 = F.dual(2, (nameOrNode, initialScope) => {
  return O.fromNullable(ASTUtils.findVariable(initialScope, nameOrNode));
});

// src/get-child-scopes.ts
function getChidScopes(scope, prev = []) {
  return [scope, ...scope.childScopes.reduce((acc, prev2) => getChidScopes(prev2, [...acc, prev2]), prev)];
}
function getScope(node, scopeManager) {
  const { type, parent } = node;
  const inner = type !== AST_NODE_TYPES.Program;
  const scope = scopeManager.acquire(node, inner);
  if (scope) return scope;
  if (parent) return getScope(parent, scopeManager);
  return null;
}
function getStaticValue2(node, initialScope) {
  return F.pipe(
    O.fromNullable(ASTUtils.getStaticValue(node, initialScope)),
    O.map((v) => v.value)
  );
}
function getVariableDeclaratorID(node, prev) {
  switch (true) {
    case (node.type === AST_NODE_TYPES.VariableDeclarator && node.init === prev):
      return O.some(node.id);
    case (node.type === AST_NODE_TYPES.AssignmentExpression && node.right === prev):
      return O.some(node.left);
    case (node.type === AST_NODE_TYPES.BlockStatement || node.type === AST_NODE_TYPES.Program || node.parent === node):
      return O.none();
    default:
      return getVariableDeclaratorID(node.parent, node);
  }
}
function getVariableNode(at) {
  return (variable) => {
    return F.pipe(
      O.some(variable),
      O.flatMapNullable((v) => v.defs.at(at)),
      O.flatMapNullable((def) => {
        switch (true) {
          case (def.type === DefinitionType.FunctionName && def.node.type === AST_NODE_TYPES.FunctionDeclaration):
            return def.node;
          case (def.type === DefinitionType.ClassName && def.node.type === AST_NODE_TYPES.ClassDeclaration):
            return def.node;
          case (def.type === DefinitionType.Parameter && AST3.isFunction(def.node)):
            return def.node;
          case ("init" in def.node && def.node.init && !("declarations" in def.node.init)):
            return def.node.init;
          default:
            return null;
        }
      })
    );
  };
}
function getVariables(initialScope) {
  let scope = initialScope;
  const variables = [...scope.variables];
  while (scope.type !== ScopeType.global) {
    scope = scope.upper;
    variables.push(...scope.variables);
  }
  return variables.reverse();
}
function isInitializedFromSource(name, source, initialScope) {
  const latestDef = O.flatMapNullable(findVariable2(name, initialScope), (v) => v.defs.at(-1));
  if (O.isNone(latestDef)) return false;
  const { node, parent } = latestDef.value;
  if (node.type === AST_NODE_TYPES.VariableDeclarator && node.init) {
    const { init } = node;
    if (init.type === AST_NODE_TYPES.MemberExpression && init.object.type === AST_NODE_TYPES.Identifier) {
      return isInitializedFromSource(init.object.name, source, initialScope);
    }
    if (init.type === AST_NODE_TYPES.Identifier) {
      return isInitializedFromSource(init.name, source, initialScope);
    }
    return F.pipe(
      getRequireExpressionArguments(init),
      O.flatMapNullable((args) => args[0]),
      O.filter(AST3.isStringLiteral),
      // check for: `require('source')` or `require('source/...')`
      O.exists((arg) => arg.value === source || arg.value.startsWith(`${source}/`))
    );
  }
  return isMatching({ type: "ImportDeclaration", source: { value: source } }, parent);
}
function getRequireExpressionArguments(node) {
  switch (true) {
    // require('source')
    case (node.type === AST_NODE_TYPES.CallExpression && node.callee.type === AST_NODE_TYPES.Identifier && node.callee.name === "require"): {
      return O.some(node.arguments);
    }
    // require('source').variable
    case node.type === AST_NODE_TYPES.MemberExpression: {
      return getRequireExpressionArguments(node.object);
    }
  }
  return O.none();
}
var thisBlockTypes = [
  AST_NODE_TYPES.FunctionDeclaration,
  AST_NODE_TYPES.FunctionExpression,
  AST_NODE_TYPES.ClassBody,
  AST_NODE_TYPES.Program
];
function isNodeValueEqual(a, b, initialScopes) {
  const [aScope, bScope] = initialScopes;
  switch (true) {
    case a === b: {
      return true;
    }
    case (a.type === AST_NODE_TYPES.Literal && b.type === AST_NODE_TYPES.Literal): {
      return a.value === b.value;
    }
    case (a.type === AST_NODE_TYPES.TemplateElement && b.type === AST_NODE_TYPES.TemplateElement): {
      return a.value.cooked === b.value.cooked;
    }
    case (a.type === AST_NODE_TYPES.Identifier && b.type === AST_NODE_TYPES.Identifier): {
      const aVar = findVariable2(a, aScope);
      const bVar = findVariable2(b, bScope);
      const aVarNode = O.flatMap(aVar, getVariableNode(0));
      const bVarNode = O.flatMap(bVar, getVariableNode(0));
      const aVarNodeParent = O.flatMapNullable(aVarNode, (n) => n.parent);
      const bVarNodeParent = O.flatMapNullable(bVarNode, (n) => n.parent);
      const aDef = O.flatMapNullable(aVar, (v) => v.defs.at(0));
      const bDef = O.flatMapNullable(bVar, (v) => v.defs.at(0));
      const aDefParentParent = O.flatMapNullable(aDef, (d) => d.parent?.parent);
      const bDefParentParent = O.flatMapNullable(bDef, (d) => d.parent?.parent);
      switch (true) {
        case (O.exists(aVarNodeParent, AST3.is(AST_NODE_TYPES.CallExpression)) && O.exists(bVarNodeParent, AST3.is(AST_NODE_TYPES.CallExpression)) && O.exists(aVarNode, AST3.isFunction) && O.exists(bVarNode, AST3.isFunction)): {
          const hasSameCallee = F.pipe(
            O.Do,
            O.bind("aCallee", () => O.map(aVarNodeParent, (n) => n.callee)),
            O.bind("bCallee", () => O.map(bVarNodeParent, (n) => n.callee)),
            O.exists(({ aCallee, bCallee }) => AST3.isNodeEqual(aCallee, bCallee))
          );
          if (!hasSameCallee) return false;
          return F.pipe(
            O.Do,
            O.bind("aParams", () => O.map(aVarNode, (n) => n.params)),
            O.bind("bParams", () => O.map(bVarNode, (n) => n.params)),
            O.let("aPos", ({ aParams }) => aParams.findIndex(AST3.isNodeEqual(a))),
            O.let("bPos", ({ bParams }) => bParams.findIndex(AST3.isNodeEqual(b))),
            O.filter(({ aPos, bPos }) => aPos !== -1 && bPos !== -1),
            O.exists(({ aPos, bPos }) => aPos === bPos)
          );
        }
        case (O.exists(aDefParentParent, AST3.is(AST_NODE_TYPES.ForOfStatement)) && O.exists(bDefParentParent, AST3.is(AST_NODE_TYPES.ForOfStatement))): {
          return F.pipe(
            O.Do,
            O.bind("aLeft", () => O.map(aDefParentParent, (d) => d.left)),
            O.bind("bLeft", () => O.map(bDefParentParent, (d) => d.left)),
            O.filter(({ aLeft, bLeft }) => aLeft.type === bLeft.type),
            O.bind("aRight", () => O.map(aDefParentParent, (d) => d.right)),
            O.bind("bRight", () => O.map(bDefParentParent, (d) => d.right)),
            O.exists(({ aRight, bRight }) => AST3.isNodeEqual(aRight, bRight))
          );
        }
        default: {
          return O.isSome(aVar) && O.isSome(bVar) && aVar.value === bVar.value;
        }
      }
    }
    case (a.type === AST_NODE_TYPES.MemberExpression && b.type === AST_NODE_TYPES.MemberExpression): {
      return AST3.isNodeEqual(a.property, b.property) && isNodeValueEqual(a.object, b.object, initialScopes);
    }
    case (a.type === AST_NODE_TYPES.ThisExpression && b.type === AST_NODE_TYPES.ThisExpression): {
      if (aScope.block === bScope.block) return true;
      return F.pipe(
        O.Do,
        O.bind("aFunction", () => AST3.findParentNode(a, AST3.isOneOf(thisBlockTypes))),
        O.bind("bFunction", () => AST3.findParentNode(b, AST3.isOneOf(thisBlockTypes))),
        O.exists(({ aFunction, bFunction }) => aFunction === bFunction)
      );
    }
    default: {
      return F.pipe(
        O.Do,
        O.bind("aStatic", () => getStaticValue2(a, aScope)),
        O.bind("bStatic", () => getStaticValue2(b, bScope)),
        O.exists(({ aStatic, bStatic }) => aStatic === bStatic)
      );
    }
  }
}
var ValueConstructionHint = {
  None: 0n,
  StrictCallExpression: 1n << 0n
};
function getValueConstruction(node, initialScope, hint = ValueConstructionHint.None) {
  switch (node.type) {
    case AST_NODE_TYPES.JSXElement:
    case AST_NODE_TYPES.JSXFragment: {
      return { kind: "JSXElement", node };
    }
    case AST_NODE_TYPES.ArrayExpression: {
      return { kind: "ArrayExpression", node };
    }
    case AST_NODE_TYPES.ObjectExpression: {
      return { kind: "ObjectExpression", node };
    }
    case AST_NODE_TYPES.ClassExpression: {
      return { kind: "ClassExpression", node };
    }
    case AST_NODE_TYPES.NewExpression: {
      return { kind: "NewExpression", node };
    }
    case AST_NODE_TYPES.FunctionExpression:
    case AST_NODE_TYPES.ArrowFunctionExpression: {
      return { kind: "FunctionExpression", node };
    }
    case AST_NODE_TYPES.CallExpression: {
      if (hint & ValueConstructionHint.StrictCallExpression) {
        return { kind: "CallExpression", node };
      }
      return { kind: "None", node };
    }
    case AST_NODE_TYPES.MemberExpression: {
      if (!("object" in node)) return { kind: "None", node };
      return getValueConstruction(node.object, initialScope, hint);
    }
    case AST_NODE_TYPES.AssignmentExpression:
    case AST_NODE_TYPES.AssignmentPattern: {
      if (!("right" in node)) return { kind: "None", node };
      return getValueConstruction(node.right, initialScope, hint);
    }
    case AST_NODE_TYPES.LogicalExpression: {
      const lvc = getValueConstruction(node.left, initialScope, hint);
      if (lvc.kind !== "None") return lvc;
      return getValueConstruction(node.right, initialScope, hint);
    }
    case AST_NODE_TYPES.ConditionalExpression: {
      const cvc = getValueConstruction(node.consequent, initialScope, hint);
      if (cvc.kind !== "None") return cvc;
      return getValueConstruction(node.alternate, initialScope, hint);
    }
    case AST_NODE_TYPES.Identifier: {
      if (!("name" in node && isString(node.name))) return { kind: "None", node };
      const construction = F.pipe(
        O.fromNullable(initialScope.set.get(node.name)),
        O.flatMap(getVariableNode(-1)),
        O.map((node2) => getValueConstruction(node2, initialScope, hint)),
        O.filter((vc) => vc.kind !== "None")
      );
      if (O.isNone(construction)) return { kind: "None", node };
      return {
        ...construction.value
      };
    }
    case AST_NODE_TYPES.Literal: {
      if ("regex" in node) {
        return { kind: "RegExpLiteral", node };
      }
      return { kind: "None", node };
    }
    default: {
      if (!("expression" in node && isObject(node.expression))) {
        return { kind: "None", node };
      }
      return getValueConstruction(node.expression, initialScope, hint);
    }
  }
}

export { ValueConstructionHint, findVariable2 as findVariable, getChidScopes, getScope, getStaticValue2 as getStaticValue, getValueConstruction, getVariableDeclaratorID, getVariableNode, getVariables, isInitializedFromSource, isNodeValueEqual };
